# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Main bdbcontrib API.

The bdbcontrib module serves a sandbox for experimental and semi-stable
features that are not yet ready for integreation to the bayeslite repository.
"""

from bql_utils import cardinality
from bql_utils import cursor_to_df
from bql_utils import describe_generator
from bql_utils import describe_generator_columns
from bql_utils import describe_generator_models
from bql_utils import describe_table
from bql_utils import df_to_table
from bql_utils import nullify
from bql_utils import table_to_df
from bql_utils import query

from diagnostic_utils import estimate_kl_divergence
from diagnostic_utils import estimate_log_likelihood

from version import __version__

def draw_crosscat(*args, **kwargs):
    import crosscat_utils
    draw_crosscat = crosscat_utils.draw_crosscat
    return draw_crosscat(*args, **kwargs)

def plot_crosscat_chain_diagnostics(*args, **kwargs):
    import crosscat_utils
    plot_crosscat_chain_diagnostics = \
        crosscat_utils.plot_crosscat_chain_diagnostics
    return plot_crosscat_chain_diagnostics(*args, **kwargs)

def barplot(*args, **kwargs):
    import plot_utils
    barplot = plot_utils.barplot
    return barplot(*args, **kwargs)

def heatmap(*args, **kwargs):
    import plot_utils
    heatmap = plot_utils.heatmap
    return heatmap(*args, **kwargs)

def histogram(*args, **kwargs):
    import plot_utils
    histogram = plot_utils.histogram
    return histogram(*args, **kwargs)

def mi_hist(*args, **kwargs):
    import plot_utils
    mi_hist = plot_utils.mi_hist
    return mi_hist(*args, **kwargs)

def pairplot(*args, **kwargs):
    import plot_utils
    pairplot = plot_utils.pairplot
    return pairplot(*args, **kwargs)

__all__ = [
    # bql_utils
        'cardinality',
        'cursor_to_df',
        'describe_generator',
        'describe_generator_columns',
        'describe_generator_models',
        'describe_table',
        'df_to_table',
        'nullify',
        'table_to_df',
        'query',
    # crosscat_utils
        'draw_crosscat',
        'plot_crosscat_chain_diagnostics',
    # diagnostic_utils
        'estimate_kl_divergence',
        'estimate_log_likelihood',
    # plot_utils
        'barplot',
        'heatmap',
        'histogram',
        'mi_hist',
        'pairplot',
    # version
        '__version__',
]



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from bdbcontrib.bql_utils import cursor_to_df
import multiprocessing as mp
from bayeslite import bayesdb_open


def _query_into_queue(query_string, queue, bdb_file):
    """
    Estimate pairwise similarity of a certain subset of the bdb according to
    query_string; place it in the multiprocessing Manager.Queue().

    For two technical reasons, this function is defined as a toplevel class and
    independently creates a bdb handle:

    1) Multiprocessing workers must be pickleable, and thus must be
       declared as toplevel functions;
    2) Multiple threads cannot access the same bdb handle, lest concurrency
       issues arise with corrupt data.

    Parameters
    ----------
    query_string : str
        Name of the query to execute, determined by estimate_similarity_mp.
    queue : multiprocessing.Manager.Queue
        Queue to place results into
    bdb_file : str
        File location of the BayesDB database. This function will
        independently open a new BayesDB handler.
    """
    bdb = bayesdb_open(pathname=bdb_file)
    res = bdb.execute(query_string)
    queue.put(cursor_to_df(res))


def _chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in xrange(0, len(l), n):
        yield l[i:i+n]


def estimate_pairwise_similarity(bdb_file, table, model, sim_table=None,
                                 cores=None, N=None, overwrite=False):
    """
    Estimate pairwise similarity from the given model, splitting processing
    across multiple processors, and save results into sim_table.

    Because called methods in this function must also open up separate BayesDB
    instances, this function accepts a BayesDB filename, rather than an actual
    bayeslite.BayesDB object.

    Parameters
    ----------
    bdb_file : str
        File location of the BayesDB database object. This function will
        handle opening the file with bayeslite.bayesdb_open.
    table : str
        Name of the table containing the raw data.
    model : str
        Name of the metamodel to estimate from.
    sim_table : str
        Name of the table to insert similarity results into. Defaults to
        table name + '_similarity'.
    cores : int
        Number of processors to use. Defaults to the number of cores as
        identified by multiprocessing.num_cores.
    N : int
        Number of rows for which to estimate pairwise similarities (so
        N^2 calculations are done). Should be used just to test small
        batches; currently, there is no control over what specific pairwise
        similarities are estimated with this parameter.
    overwrite : bool
        Whether to overwrite the sim_table if it already exists. If
        overwrite=False and the table exists, function will raise
        sqlite3.OperationalError. Default True.
    """
    bdb = bayesdb_open(pathname=bdb_file)

    if cores is None:
        cores = mp.cpu_count()

    if cores < 1:
        raise ValueError(
            "Invalid number of cores {}".format(cores)
        )

    if sim_table is None:
        sim_table = table + '_similarity'

    # Get number of occurrences in the database
    count_cursor = bdb.execute('SELECT COUNT(*) FROM {}'.format(table))
    table_count = int(cursor_to_df(count_cursor)['"COUNT"(*)'][0])
    if N is None:
        N = table_count
    elif N > table_count:
        raise ValueError(
            "Asked for N={} rows but {} rows in table".format(N, table_count)
        )

    # Calculate the size (# of similarities to compute) and
    # offset (where to start) calculation for each worker query.

    # Divide sizes evenly, and make the last job finish the remainder
    sizes = [(N * N) / cores for i in range(cores)]
    sizes[-1] += (N * N) % cores

    total = 0
    offsets = [total]
    for size in sizes[:-1]:
        total += size
        offsets.append(total)

    q_template = ('ESTIMATE SIMILARITY FROM PAIRWISE {} '.format(model) +
                  'LIMIT {} OFFSET {}')  # Format sizes/offsets later

    queries = [q_template.format(*so) for so in zip(sizes, offsets)]

    # Create the similarity table. Assumes original table has rowid column.
    # XXX: tables from verbnet bdb don't necessarily have an
    # autoincrementing primary key other than rowid (doesn't work).
    # So we ought to ask for a foreign key, but ESTIMATE SIMILARITY
    # returns numerical values rather than row names, so that code
    # would have to be changed first. For now, we eliminate
    # REFERENCE {table}(foreign_key) from the name0 and name1 col specs.
    if overwrite:
        bdb.sql_execute('DROP TABLE IF EXISTS {}'.format(sim_table))

    bdb.sql_execute('''
        CREATE TABLE {sim_table} (
            rowid0 INTEGER NOT NULL,
            rowid1 INTEGER NOT NULL,
            value DOUBLE NOT NULL
        )
    '''.format(sim_table=sim_table))

    # Define the helper which inserts data into table in batches
    def insert_into_sim(df):
        """
        Use the main thread bdb handle to successively insert results of
        ESTIMATEs into the table.
        """
        # Because the bayeslite implementation of sqlite3 doesn't allow
        # inserts of > 500 rows at a time (else sqlite3.OperationalError),
        # we split the list into chunks of size 500 and perform multiple
        # insert statements.
        rows = map(list, df.values)
        rows_str = ['({})'.format(','.join(map(str, r))) for r in rows]
        for rows_chunk in _chunks(rows_str, 500):
            insert_str = '''
                INSERT INTO {} (rowid0, rowid1, value) VALUES {};
            '''.format(sim_table, ','.join(rows_chunk))
            bdb.sql_execute(insert_str)

    pool = mp.Pool(processes=cores)

    manager = mp.Manager()
    queue = manager.Queue()

    for query in queries:
        pool.apply_async(
            _query_into_queue, args=(query, queue, bdb_file)
        )

    # Close pool and wait for processes to finish
    # FIXME: This waits for all processes to finish before inserting
    # into the table, which means that memory usage is potentially very
    # high!
    pool.close()
    pool.join()

    # Process returned results
    while not queue.empty():
        df = queue.get()
        insert_into_sim(df)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Because this test may indirectly import pylab, which would peg the
# matplotlib backend, which might prevent a later test from drawing
# pictures headless.  &#&*%(@#^&!@.
import matplotlib
matplotlib.use("Agg")

import pytest
import tempfile

import bayeslite

from bdbcontrib import bql_utils
from bdbcontrib import shell_utils

csv_data = '''
id,one,two,three,four
0,1,3,1,four
1,5,4,5,four
2,1,5,2,three
3,4,4,3,five
4,0,3,1,one
5,0,1,2,three
6,1,1,1,four
7,3,3,1,one
8,2,1,5,two
9,0,2,0,one
'''.lstrip()


csv_data_nan = '''
id,one,two,three,four
0,NaN,3,NaN,four
1,5,4,5,four
2,NaN,5,2,three
3,4,4,3,five
4,0,3,NaN,NaN
5,0,NaN,2,three
6,NaN,NaN,NaN,four
7,3,3,NaN,NaN
8,2,NaN,5,two
9,0,2,0,NaN
'''.lstrip()

csv_data_empty = '''
id,one,two,three,four
0,,3,,four
1,5,4,5,four
2,,5,2,three
3,4,4,3,five
4,0,3,,""
5,0,,2,three
6,,,,four
7,3,3,,""
8,2,,5,two
9,0,2,0,""
'''.lstrip()

csv_data_999 = '''
id,one,two,three,four
0,999,3,999,four
1,5,4,5,four
2,999,5,2,three
3,4,4,3,five
4,0,3,999,999
5,0,999,2,three
6,999,999,999,four
7,3,3,999,999
8,2,999,5,two
9,0,2,0,999
'''.lstrip()


@pytest.mark.parametrize(
    "data, value, num_nulls_expected",
    [[csv_data, 'NaN', (0, 0, 0, 0,)],
     [csv_data_nan, 'NaN', (3, 3, 4, 3,)],
     [csv_data_999, '999', (3, 3, 4, 3,)],
     [csv_data_empty, '', (3, 3, 4, 3,)],
     [csv_data_nan, '999', (0, 0, 0, 0,)]])
def test_nullify_no_missing(data, value, num_nulls_expected):
    with tempfile.NamedTemporaryFile() as temp:
        temp.write(data)
        temp.seek(0)
        with bayeslite.bayesdb_open() as bdb:
            bayeslite.bayesdb_read_csv_file(bdb, 't', temp.name, header=True,
                                            create=True)
            bql_utils.nullify(bdb, 't', value)

            c = bdb.execute('SELECT COUNT(*) FROM t WHERE one IS NULL;')
            assert c.fetchvalue() == num_nulls_expected[0]

            c = bdb.execute('SELECT COUNT(*) FROM t WHERE two IS NULL;')
            assert c.fetchvalue() == num_nulls_expected[1]

            c = bdb.execute('SELECT COUNT(*) FROM t WHERE three IS NULL;')
            assert c.fetchvalue() == num_nulls_expected[2]

            c = bdb.execute('SELECT COUNT(*) FROM t WHERE four IS NULL;')
            assert c.fetchvalue() == num_nulls_expected[3]

def test_cursor_to_df():
    with bayeslite.bayesdb_open() as bdb:
        bql_utils.cursor_to_df(bdb.execute('select * from sqlite_master'))
        bql_utils.cursor_to_df(bdb.execute('select * from sqlite_master'
                ' where 0 = 1'))


def test_is_plotting_command():
    cmd1 = '.heatmap ESTIMATE PAIRWISE DEPENDENCE PROBABILITY FROM t; -f z.png'
    cmd2 = '.show SELECT a, b FROM t LIMIT 10; --no-contour'
    cmd3 = '.histogram SELECT a FROM t;'
    cmd4 = '.help histogram'
    cmd5 = 'SELECT a FROM t;'
    cmd6 = '-- .show is a plotting function'

    assert shell_utils.is_plotting_command(cmd1)
    assert shell_utils.is_plotting_command(cmd2)
    assert shell_utils.is_plotting_command(cmd3)
    assert not shell_utils.is_plotting_command(cmd4)
    assert not shell_utils.is_plotting_command(cmd5)
    assert not shell_utils.is_plotting_command(cmd6)


def test_clean_cmd_filename():
    output_dir = 'foo/baz'
    fignum = 0

    cmd1 = '.histogram SELECT a FROM t;'
    cmd2 = '.histogram SELECT a FROM t; -f pic.png'
    cmd1_expected = '.histogram SELECT a FROM t; --filename foo/baz/fig_0.png'

    cmd1_ud, _ = shell_utils.clean_cmd_filename(cmd1, fignum, output_dir)
    assert cmd1_ud == cmd1_expected
    with pytest.raises(ValueError):
        shell_utils.clean_cmd_filename(cmd2, fignum, output_dir)
    with pytest.raises(ValueError):
        shell_utils.clean_cmd_filename(cmd1_expected, fignum, output_dir)


def test_is_comment():
    assert not shell_utils.is_comment('.heatmap ESTIMATE PAIRWISE FOO;')
    assert shell_utils.is_comment('-- is a comment')
    assert shell_utils.is_comment('--is a comment')
    assert shell_utils.is_comment('---is a comment')

    # XXX: There is no support for inline comments in BQL scripts
    assert not shell_utils.is_comment(' -- not a comment')
    assert not shell_utils.is_comment('.show SELECT a, b FROM t;'
        ' --no-contour')


def test_is_dot_command():
    assert not shell_utils.is_dot_command('-- .show SELECT a, b FROM t;')
    assert not shell_utils.is_dot_command('SELECT ".show" FROM t;')
    assert shell_utils.is_dot_command('.show SELECT a, b FROM t;')

@pytest.mark.parametrize(
    "data, cols, cardinalities_expected",
    [[csv_data, None, [10, 6, 5, 5, 5]],
     [csv_data_nan, None, [10, 6, 5, 5, 5]],
     [csv_data_999, None, [10, 6, 5, 5, 5]],
     [csv_data_empty, None, [10, 6, 5, 5, 5]],
     [csv_data, ['id'], [10]],
     [csv_data, ['two', 'id'], [5, 10]],
     ])
def test_cardinality(data, cols, cardinalities_expected):
    with tempfile.NamedTemporaryFile() as temp:
        temp.write(data)
        temp.seek(0)
        with bayeslite.bayesdb_open() as bdb:
            bayeslite.bayesdb_read_csv_file(bdb, 't', temp.name, header=True,
                                            create=True)
            cards = bql_utils.cardinality(bdb, 't', cols)
            for c in cards:
                assert 2 == len(c)
                assert c[0] in ('id', 'one', 'two', 'three', 'four')
            assert cardinalities_expected == [c[1] for c in cards]



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import print_function

# matplotlib needs to set the backend before anything else gets to.
import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt

from collections import namedtuple
from contextlib import contextmanager
from io import BytesIO
from string import ascii_lowercase  # pylint: disable=deprecated-module
from textwrap import dedent
import bayeslite
import numpy as np
import os
import pandas as pd
import pytest
import random
import re
import sys
import tempfile
import test_plot_utils

from bayeslite.read_csv import bayesdb_read_csv
from bayeslite.loggers import BqlLogger
from bdbcontrib import recipes

testvars = {'dataset': None, 'input_df': None}

import multiprocessing
import time
@contextmanager
def ensure_timeout(delay, target):
    proc = multiprocessing.Process(target=target)
    proc.start()
    proc.join(delay)
    assert not proc.is_alive()
    # proc.terminate()
    # proc.join()


class MyTestLogger(BqlLogger):
    def __init__(self, verbose=pytest.config.option.verbose):
        self.calls = []
        self.verbose = verbose
    def info(self, msg_format, *values):
        if self.verbose:
            print('INFO: '+msg_format, *values)
        self.calls.append(('info', msg_format, values))
    def warn(self, msg_format, *values):
        if self.verbose:
            print('WARN: '+msg_format, *values)
        self.calls.append(('warn', msg_format, values))
    def plot(self, suggested_name, figure):
        if self.verbose:
            plt.show()
        self.calls.append(('plot', suggested_name, figure))
    def result(self, msg_format, *values):
        if self.verbose:
            print('RSLT: '+msg_format, *values)
        self.calls.append(('result', msg_format, values))


@contextmanager
def prepare():
    #import pytest
    #pytest.set_trace()
    if testvars['dataset'] is None:
        (df, csv_data) = test_plot_utils.dataset(40)
        tempd = tempfile.mkdtemp(prefix="bdbcontrib-test-recipes")
        csv_path = os.path.join(tempd, "data.csv")
        with open(csv_path, "w") as csv_f:
            csv_f.write(csv_data.getvalue())
        bdb_path = os.path.join(tempd, "data.bdb")
        name = ''.join(random.choice(ascii_lowercase) for _ in range(32))
        dts = recipes.quickstart(name=name,
                                 csv_path=csv_path, bdb_path=bdb_path,
                                 logger=MyTestLogger(),
                                 session_capture_name="test_recipes.py")
        ensure_timeout(10, lambda: dts.analyze(models=10, iterations=20))
        testvars['dataset'] = dts
        testvars['input_df'] = df

    yield testvars['dataset'], testvars['input_df']

def test_analyze_and_analysis_status_and_reset():
    with prepare() as (dts, _df):
        resultdf = dts.analysis_status()
        assert 'iterations' == resultdf.index.name, repr(resultdf)
        assert 'count of models' == resultdf.columns[0], repr(resultdf)
        assert 1 == len(resultdf), repr(resultdf)
        assert 10 == resultdf.ix[20, 0], repr(resultdf)

        resultdf = dts.analyze(models=11, iterations=1)
        assert 'iterations' == resultdf.index.name, repr(resultdf)
        assert 'count of models' == resultdf.columns[0], repr(resultdf)
        dts.logger.result(str(resultdf))
        assert 2 == len(resultdf), repr(resultdf)
        assert 10 == resultdf.ix[21, 0], repr(resultdf)
        assert 1 == resultdf.ix[1, 0], repr(resultdf)

        dts.reset()
        resultdf = dts.analysis_status()
        assert 'iterations' == resultdf.index.name, repr(resultdf)
        assert 'count of models' == resultdf.columns[0], repr(resultdf)
        assert 0 == len(resultdf), repr(resultdf)
        ensure_timeout(10, lambda: dts.analyze(models=10, iterations=20))
        resultdf = dts.analysis_status()
        assert 'iterations' == resultdf.index.name, repr(resultdf)
        assert 'count of models' == resultdf.columns[0], repr(resultdf)
        assert 1 == len(resultdf), repr(resultdf)
        assert 10 == resultdf.ix[20, 0], repr(resultdf)

        # This is the only test that needs the files anymore (for reset),
        # so now that we're done, clean those up. The rest of the tests can
        # happen in any order based on the in-memory bdb.
        import shutil
        shutil.rmtree(os.path.dirname(dts.csv_path))

def test_q():
    with prepare() as (dts, df):
        resultdf = dts.query('SELECT COUNT(*) FROM %t;')
        #resultdf.to_csv(sys.stderr, header=True)
        assert 1 == len(resultdf)
        assert 1 == len(resultdf.columns)
        assert '"COUNT"(*)' == resultdf.columns[0]
        assert len(df) == resultdf.iloc[0, 0]
        resultdf = dts.query(dedent('''\
            ESTIMATE DEPENDENCE PROBABILITY OF
            floats_1 WITH categorical_1 BY %g'''))
        resultdf.to_csv(sys.stderr, header=True)

def test_quick_describe_columns_and_column_type():
    with prepare() as (dts, _df):
        resultdf = dts.query('SELECT * from %t LIMIT 1')
        assert ['index', 'floats_1', 'categorical_1', 'categorical_2',
                'few_ints_3', 'floats_3', 'many_ints_4', 'skewed_numeric_5',
                ] == list(resultdf.columns)
        resultdf = dts.quick_describe_columns()
        expected = {'floats_1': 'numerical',
                    'categorical_1': 'categorical',
                    'categorical_2': 'categorical',
                    'few_ints_3': 'categorical',
                    'floats_3': 'numerical',
                    'many_ints_4': 'numerical',
                    'skewed_numeric_5': r'(categorical|numerical)'}
        for (column, expected_type) in expected.iteritems():
            stattype = resultdf[resultdf['name'] == column]['stattype'].iloc[0]
            assert re.match(expected_type, stattype), column
            assert re.match(expected_type, dts.column_type(column))

def test_heatmap():
    import sys
    with prepare() as (dts, _df):
        dts.logger.calls = []
        deps = dts.query('ESTIMATE DEPENDENCE PROBABILITY'
                         ' FROM PAIRWISE COLUMNS OF %g')
        dts.heatmap(deps, plotfile='foobar.png')
        plots = [c for c in dts.logger.calls if c[0] == 'plot']
        assert 1 == len(plots)
        thecall = plots[0]
        assert ('plot', 'foobar.png') == thecall[:2]

        dts.logger.calls = []
        dts.heatmap(deps, plotfile='foobar.png',
                    selectors=
                    {'A-E': lambda x: bool(re.search(r'^[a-eA-E]', x[0])),
                     'F-W': lambda x: bool(re.search(r'^[f-wF-W]', x[0])),
                     'X-Z': lambda x: bool(re.search(r'^[x-zX-Z]', x[0]))})
        # NOTE: columns are: 'floats_1', 'categorical_1', 'categorical_2',
        #     'few_ints_3', 'floats_3', 'many_ints_4', 'skewed_numeric_5',
        # So there are 2 in A-E, 5 in F-W, and 0 in X-Z.
        # So we should produce 3 plots: AE vs. AE, AE vs. FW, and FW vs. FW.
        plots = [c for c in dts.logger.calls if c[0] == 'plot']
        assert 4 == len(plots)
        names = [p[1] for p in plots]
        assert 'A-E.A-E.foobar.png' in names
        assert 'A-E.F-W.foobar.png' in names
        assert 'F-W.A-E.foobar.png' in names
        assert 'F-W.F-W.foobar.png' in names

def test_explore_cols():
    with prepare() as (dts, _df):
        dts.logger.calls = []
        try:
            dts.quick_explore_cols([])
            assert False, "Should raise a ValueError because empty columns."
        except ValueError:
            pass
        try:
            dts.quick_explore_cols(['float_1'])
            assert False, "Should raise a ValueError because only one column."
        except ValueError:
            pass

        dts.quick_explore_cols(['floats_1', 'categorical_1'])
        call_types = pd.DataFrame([call[0] for call in dts.logger.calls])
        call_counts = call_types.iloc[:,0].value_counts()
        assert call_counts['result'] > 0
        assert call_counts['plot'] > 0
        assert 'warn' not in call_counts

def test_pairplot():
    with prepare() as (dts, _df):
        dts.logger.calls = []
        try:
            dts.pairplot([])
            assert False, "Should raise a ValueError because empty columns."
        except ValueError:
            pass

        pt = dts.pairplot(
            ['floats_1', 'floats_3', 'many_ints_4', 'categorical_1'],
            colorby='categorical_2')
        call_types = pd.DataFrame([call[0] for call in dts.logger.calls])
        call_counts = call_types.iloc[:,0].value_counts()
        assert call_counts['plot'] > 0
        assert 'warn' not in call_counts

def test_similar_rows():
    with prepare() as (dts, _df):
        dts.logger.calls = []
        the_real_queryfn = dts.query
        count = [0]
        q_calls = []
        results = [pd.DataFrame([[0]]),  # Count of matching rows.
                   pd.DataFrame([[2]]),  # Count of matching rows.
                   pd.DataFrame([[1]]),  # Count of matching rows.
                   None,                 # Call with temp table creation.
                   'similar rows',       # Similarity result.
                   ]
        def my_test_queryfn(fmt_str, *args):
            q_calls.append((fmt_str, args))
            count[0] += 1
            return results[count[0] - 1]
        try:
            dts.query = my_test_queryfn
            try:
                dts.quick_similar_rows(identify_row_by={})
                assert False, "Expected death due to no rows."
            except NotImplementedError:
                pass
            assert 'SELECT COUNT(*)' == q_calls[-1][0][:15]
            try:
                dts.quick_similar_rows(identify_row_by={})
                assert False, "Expected death due to two many rows."
            except NotImplementedError:
                pass
            assert 'SELECT COUNT(*)' == q_calls[-1][0][:15]
            assert 'similar rows' == dts.quick_similar_rows(identify_row_by={
                'a':'b', 'c': 'd'})
            assert 'SELECT COUNT(*)' == q_calls[-3][0][:15]
            assert 'CREATE TEMP TABLE' == q_calls[-2][0][:17]
            assert re.search(r'''"a" = [?]\s+and\s+"c" = [?] ''',
                             q_calls[-2][0])
            assert (['b', 'd'],) == q_calls[-2][1]
            assert 'SELECT * FROM' == q_calls[-1][0][:13]
        finally:
            dts.query = the_real_queryfn


        call_types = pd.DataFrame([call[0] for call in dts.logger.calls])
        if len(call_types) > 0:
            call_counts = call_types.iloc[:,0].value_counts()
            assert 'warn' not in call_counts



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from contextlib import contextmanager
import copy
from textwrap import wrap

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

import bayeslite.core

import bdbcontrib.bql_utils as bqlu

###############################################################################
###                                   PUBLIC                                ###
###############################################################################

def mi_hist(bdb, generator, col1, col2, num_samples=1000, bins=5):
    """Plot histogram of mutual information over generator's models.

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        Active BayesDB instance.
    generator : str
        Name of the generator to compute MI(col1;col2)
    col1, col2 : str
        Name of the columns to compute MI(col1;col2)
    num_samples : int, optional
        Number of samples to use in the Monte Carlo estimate of MI.
    bins : int, optional
        Number of bins in the histogram.

    Returns
    -------
    figure : matplotlib.figure.Figure
    """
    generator_id = bayeslite.core.bayesdb_get_generator(bdb, generator)
    bql = '''
        SELECT COUNT(modelno) FROM bayesdb_generator_model
            WHERE generator_id = ?
    '''
    counts = bdb.execute(bql, (generator_id,))
    num_models = counts.fetchvalue()

    figure, ax = plt.subplots(figsize=(6, 6))

    mis = []
    for modelno in range(num_models):
        bql = '''
            ESTIMATE MUTUAL INFORMATION OF {} WITH {} USING {} SAMPLES FROM {}
                USING MODEL {} LIMIT 1
        '''.format(col1, col2, num_samples, generator, modelno)
        cursor = bdb.execute(bql)
        mutual_information = cursor.fetchvalue()
        mis.append(mutual_information)
    ax.hist(mis, bins, normed=True)
    ax.set_xlabel('Mutual Information')
    ax.set_ylabel('Density')
    ax.set_title('Mutual information of {} with {}'.format(col1, col2))

    return figure

def heatmap(bdb, bql=None, df=None, **kwargs):
    """Plot clustered heatmap of pairwise matrix.

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        Active BayesDB instance.
    bql : str
        The BQL to run and plot. Must be a PAIRWISE BQL query if specified.
        One of bql or df must be specified.
    df : pandas.DataFrame(columns=['generator_id', 'name0', 'name1', 'value'])
        If bql is not specified, take data from here.

    **kwargs : dict
        Passed to zmatrix: vmin, vmax, row_ordering, col_ordering

    Returns
    -------
    clustermap: seaborn.clustermap
    """
    assert bql is not None or df is not None
    assert bql is None or df is None
    if bql is not None:
        df = bqlu.cursor_to_df(bdb.execute(bql))
    df.fillna(0, inplace=True)
    return zmatrix(df, **kwargs)

def selected_heatmaps(bdb, selectors, bql=None, df=None, **kwargs):
    """Yield heatmaps of pairwise matrix, broken up according to selectors.

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        Active BayesDB instance.
    selectors : [lambda name --> bool]
        Rather than plot the full NxN matrix all together, make separate plots
        for each combination of these selectors, plotting them in sequence.
        If selectors are specified, yields clustermaps, which caller is
        responsible for showing or saving, and then closing.
    bql : str
        The BQL to run and plot. Must be a PAIRWISE BQL query if specified.
        One of bql or df must be specified.
    df : pandas.DataFrame(columns=['generator_id', 'name0', 'name1', 'value'])
        If bql is not specified, then take data from here instead.
    **kwargs : dict
        Passed to zmatrix: vmin, vmax, row_ordering, col_ordering

    Yields
    ------
    The triple (clustermap, selector1, selector2).  It is recommended that
    caller keep a dict of these functions to names to help identify each one.
    """
    # Cannot specify neither or both.
    assert bql is not None or df is not None
    assert bql is None or df is None
    if bql is not None:
        df = bqlu.cursor_to_df(bdb.execute(bql))
    df.fillna(0, inplace=True)
    for n0selector in selectors:
        n0selection = df.iloc[:, 1].map(n0selector)
        for n1selector in selectors:
            n1selection = df.iloc[:, 2].map(n1selector)
            this_block = df[n0selection & n1selection]
            if len(this_block) > 0:
                yield (zmatrix(this_block, vmin=0, vmax=1, **kwargs),
                       n0selector, n1selector)

def pairplot(bdb, bql, generator_name=None, show_contour=False, colorby=None,
        show_missing=False, show_full=False):
    """Plot array of plots for all pairs of columns.

    Plots continuous-continuous pairs as scatter (optional KDE contour).
    Plots continuous-categorical pairs as violinplot.
    Plots categorical-categorical pairs as heatmap.

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        Active BayesDB instance.
    bql : str
        The BQL to run and pairplot.
    generator_name : str, optional
        The name of generator; explicitly passing in provides optimizations.
    show_contour : bool, optional
        If True, KDE contours are plotted on top of scatter plots
        and histograms.
    show_missing : bool, optional
        If True, rows with one missing value are plotted as lines on scatter
        plots.
    colorby : str, optional
        Name of a column to use to color data points in histograms and scatter
        plots.
    show_full : bool, optional
        Show full pairwise plots, rather than only lower triangular plots.

    Returns
    -------
    figure : matplotlib.figure.Figure
    """
    df = bqlu.cursor_to_df(bdb.execute(bql))
    figure = _pairplot(df, bdb=bdb, generator_name=generator_name,
        show_contour=show_contour, colorby=colorby, show_missing=show_missing,
        show_full=show_full)
    figure.tight_layout()
    inches = len(df.columns) * 4
    figure.set_size_inches((inches, inches))

    return figure


def histogram(bdb, bql, bins=15, normed=None):
    """Plot histogram of one- or two-column table.

    If two-column, subdivide the first column according to labels in
    the second column

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        Active BayesDB instance.
    bql : str
        The BQL to run and histogram.
    bins : int, optional
        Number of bins in the histogram.
    normed : bool, optional
        Normalize the histograms?

    Returns
    ----------
    figure: matplotlib.figure.Figure
    """
    df = bqlu.cursor_to_df(bdb.execute(bql))
    figure = comparative_hist(df, bdb=bdb, nbins=bins, normed=normed)

    return figure


def barplot(bdb, bql):
    """Plot bar-plot of query giving categories and heights.

    First column specifies names; second column specifies heights.

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        Active BayesDB instance.
    bql : str
        The BQL to run and histogram. Must have a two-column result.

    Returns
    ----------
    figure: matplotlib.figure.Figure
    """
    df = bqlu.cursor_to_df(bdb.execute(bql))
    if df.shape[1] != 2:
        raise ValueError('Need two columns of output from SELECT for barplot.')
    height_inches = df.shape[0] / 2.0
    figure, ax = plt.subplots(figsize=(height_inches, 5))

    ax.bar([x - .5 for x in range(df.shape[0])], df.ix[:, 1].values,
        color='#333333', edgecolor='#333333')
    ax.set_xticks(range(df.shape[0]))
    ax.set_xticklabels(df.ix[:, 0].values, rotation=90)
    ax.set_xlim([-1, df.shape[0] - .5])
    ax.set_ylabel(df.columns[1])
    ax.set_xlabel(df.columns[0])
    ax.set_title('\n    '.join(wrap(bql, 80)))

    return figure


###############################################################################
###                              INTERNAL                                   ###
###############################################################################

MODEL_TO_TYPE_LOOKUP = {
    'normal_inverse_gamma': 'numerical',
    'symmetric_dirichlet_discrete': 'categorical',
}


def rotate_tick_labels(ax, axis='x', rotation=90):
    if axis.lower() == 'x':
        _, labels = ax.get_xticks()
    elif axis.lower() == 'y':
        _, labels = ax.get_yticks()
    else:
        raise ValueError('axis must be x or y')
    plt.setp(labels, rotation=rotation)


def gen_collapsed_legend_from_dict(hl_colors_dict, loc=0, title=None,
        fontsize='medium', wrap_threshold=1000):
    """Creates a legend with entries grouped by color.

    For example, if a plot has multiple labels associated with the same color
    line, instead of generating a legend entry for each label, labels with the
    same colored line will be collapsed into longer, comma-separated labels.

    Parameters
    ----------
    hl_colors_dict : dict
        A dict of label, color pairs. Colors can be strings e.g. 'deeppink' or
        rgb or rgba tuples.
    loc : matplotlib compatible
        any matpltotlib-compbatible legend location identifier
    title : str
        legend title
    fontsize : int
        legend entry and title fontsize
    wrap_threshold : int
        max number of characters before wordwrap

    Returns
    -------
    legend : matplotlib.legend
    """
    if not isinstance(hl_colors_dict, dict):
        raise TypeError("hl_colors_dict must be a dict")

    colors = list(set(hl_colors_dict.values()))
    from collections import defaultdict
    collapsed_dict = defaultdict(list)

    for label, color in hl_colors_dict.iteritems():
        collapsed_dict[color].append(str(label))

    for color in collapsed_dict.keys():
        collapsed_dict[color] = "\n".join(wrap(", ".join(
            sorted(collapsed_dict[color])), wrap_threshold))

    legend_artists = []
    legend_labels = []
    for color, label in collapsed_dict.iteritems():
        legend_artists.append(plt.Line2D((0, 1), (0, 0), color=color, lw=3))
        legend_labels.append(label)

    legend = plt.legend(legend_artists, legend_labels, loc=loc, title=title,
        fontsize=fontsize)

    return legend


def get_bayesdb_col_type(column_name, df_column, bdb=None,
                         generator_name=None):
    # If column_name is a column label (not a short name!) then the modeltype
    # of the column will be returned otherwise we guess.

    if isinstance(df_column, pd.DataFrame):
        raise TypeError('Multiple columns in the query result have the same '
            'name (%s).' % (column_name,))

    def guess_column_type(df_column):
        pd_type = df_column.dtype
        if pd_type is str:
            return 'categorical'
        else:
            if len(df_column.unique()) < 30:
                return 'categorical'
            else:
                return 'numerical'

    if bdb is not None and generator_name is not None:
        try:
            coltype = bqlu.get_column_stattype(bdb, generator_name, column_name)
            # XXX: Force cyclic -> numeric because there is no need to plot
            # cyclic data any differently until we implement rose plots. See
            # http://matplotlib.org/examples/pie_and_polar_charts/polar_bar_demo.html
            # for an example.
            if coltype.lower() == 'cyclic':
                coltype = 'numerical'
            return coltype
        except IndexError:
            return guess_column_type(df_column)
        except Exception as err:
            print 'Unexpected exception: {}'.format(err)
            raise err
    else:
        return guess_column_type(df_column)


def conv_categorical_vals_to_numeric(data_srs):
    # TODO: get real valuemap from btable
    unique_vals = sorted(data_srs.unique().tolist())
    lookup = dict(zip(unique_vals, range(len(unique_vals))))
    values = data_srs.values.tolist()
    for i, x in enumerate(values):
        values[i] = lookup[x]
    return np.array(values, dtype=float), unique_vals, lookup


# FIXME: STUB
def prep_plot_df(data_df, var_names):
    return data_df[list(var_names)]

def drop_inf_and_nan(np_Series):
    return np_Series.replace([-np.inf, np.inf], np.nan).dropna()

def do_hist(data_srs, **kwargs):
    ax = kwargs.get('ax', None)
    bdb = kwargs.get('bdb', None)
    dtype = kwargs.get('dtype', None)
    generator_name = kwargs.get('generator_name', None)
    colors = kwargs.get('colors', None)

    if dtype is None:
        dtype = get_bayesdb_col_type(data_srs.columns[0], data_srs, bdb=bdb,
            generator_name=generator_name)

    if ax is None:
        ax = plt.gca()

    if len(data_srs.shape) > 1:
        if colors is None and data_srs.shape[1] != 1:
            raise ValueError('If a dummy column is specified, colors must also'
                ' be specified.')

    data_srs = data_srs.dropna()

    if dtype == 'categorical':
        vals, uvals, _ = conv_categorical_vals_to_numeric(data_srs.ix[:, 0])
        if colors is not None:
            color_lst = []
            stacks = []
            for val, color in colors.iteritems():
                subval = vals[data_srs.ix[:, 1].values == val]
                color_lst.append(color)
                stacks.append(subval)
            ax.hist(stacks, bins=len(uvals), color=color_lst, alpha=.9,
                histtype='barstacked', rwidth=1.0)
        else:
            ax.hist(vals, bins=len(uvals))
        ax.set_xticks(range(len(uvals)))
        ax.set_xticklabels(uvals)
    else:
        do_kde = True
        if colors is not None:
            for val, color in colors.iteritems():
                subdf = data_srs.loc[data_srs.ix[:, 1] == val]
                sns.distplot(drop_inf_and_nan(subdf.ix[:, 0]),
                             kde=do_kde, ax=ax, color=color)
        else:
            sns.distplot(drop_inf_and_nan(data_srs), kde=do_kde, ax=ax)

    return ax


def do_heatmap(plot_df, unused_vartypes, **kwargs):
    ax = kwargs.get('ax', None)

    plot_df = plot_df.dropna()
    if ax is None:
        ax = plt.gca()

    vals_x, uvals_x, _ = conv_categorical_vals_to_numeric(plot_df.ix[:, 0])
    vals_y, uvals_y, _ = conv_categorical_vals_to_numeric(plot_df.ix[:, 1])

    bins_x = len(uvals_x)
    bins_y = len(uvals_y)

    hst, _, _ = np.histogram2d(vals_y, vals_x, bins=[bins_y, bins_x])
    cmap = 'PuBu'
    ax.matshow(hst, aspect='auto', origin='lower', cmap=cmap)
    ax.grid(b=False)
    ax.set_xticks(range(bins_x))
    ax.set_yticks(range(bins_y))
    ax.set_xticklabels(uvals_x)
    ax.set_yticklabels(uvals_y)
    ax.xaxis.set_tick_params(labeltop='off', labelbottom='on')
    return ax


def do_violinplot(plot_df, vartypes, **kwargs):
    ax = kwargs.get('ax', None)
    colors = kwargs.get('colors', None)

    dummy = plot_df.shape[1] == 3

    plot_df = plot_df.dropna()
    if ax is None:
        ax = plt.gca()
    assert vartypes[0] != vartypes[1]
    vert = vartypes[1] == 'numerical'
    plot_df = copy.deepcopy(plot_df.dropna())
    if vert:
        groupby = plot_df.columns[0]
    else:
        groupby = plot_df.columns[1]

    _, unique_vals, _ = conv_categorical_vals_to_numeric(plot_df[groupby])

    unique_vals = np.sort(unique_vals)
    n_vals = len(plot_df[groupby].unique())
    if dummy:
        sub_vals = np.sort(plot_df[groupby].unique())
        axis = sns.violinplot(
            x=plot_df.columns[0],
            y=plot_df.columns[1],
            data=plot_df,
            order=sub_vals, hue=plot_df.columns[2],
            names=sub_vals, ax=ax, orient=("v" if vert else "h"),
            palette=colors, inner='quartile')
        axis.legend_ = None
    else:
        sns.violinplot(
            x=plot_df.columns[0],
            y=plot_df.columns[1],
            data=plot_df,
            order=unique_vals, names=unique_vals, ax=ax,
            orient=("v" if vert else "h"),
            color='SteelBlue', inner='quartile')

    if vert:
        ax.set_xlim([-.5, n_vals-.5])
        ax.set_xticks(np.arange(n_vals))
        ax.set_xticklabels(unique_vals)
    else:
        ax.set_ylim([-.5, n_vals-.5])
        ax.set_yticks(np.arange(n_vals))
        ax.set_yticklabels(unique_vals)

    return ax


def do_kdeplot(plot_df, unused_vartypes, **kwargs):
    # XXX: kdeplot is not a good choice for small amounts of data because
    # it uses a kernel density estimator to crease a smooth heatmap. On the
    # other hand, scatter plots are uninformative given lots of data---the
    # points get jumbled up. We may just want to set a threshold (N=100)?

    ax = kwargs.get('ax', None)
    show_contour = kwargs.get('show_contour', False)
    colors = kwargs.get('colors', None)
    show_missing = kwargs.get('show_missing', False)

    if ax is None:
        ax = plt.gca()

    xlim = [plot_df.ix[:, 0].min(), plot_df.ix[:, 0].max()]
    ylim = [plot_df.ix[:, 1].min(), plot_df.ix[:, 1].max()]

    dummy = plot_df.shape[1] == 3

    null_rows = plot_df[plot_df.isnull().any(axis=1)]
    df = plot_df.dropna()

    if not dummy:
        plt.scatter(df.values[:, 0], df.values[:, 1], alpha=.5,
            color='steelblue', zorder=2)
        # plot nulls
        if show_missing:
            nacol_x = null_rows.ix[:, 0].dropna()
            for x in nacol_x.values:
                plt.plot([x, x], ylim, color='crimson', alpha=.2, lw=1,
                    zorder=1)
            nacol_y = null_rows.ix[:, 1].dropna()
            for y in nacol_y.values:
                plt.plot(xlim, [y, y], color='crimson', alpha=.2, lw=1,
                    zorder=1)
    else:
        assert isinstance(colors, dict)
        for val, color in colors.iteritems():
            subdf = df.loc[df.ix[:, 2] == val]
            plt.scatter(subdf.values[:, 0], subdf.values[:, 1], alpha=.5,
                color=color, zorder=2)
            subnull = null_rows.loc[null_rows.ix[:, 2] == val]
            if show_missing:
                nacol_x = subnull.ix[:, 0].dropna()
                for x in nacol_x.values:
                    plt.plot([x, x], ylim, color=color, alpha=.3, lw=2,
                        zorder=1)
                nacol_y = subnull.ix[:, 1].dropna()
                for y in nacol_y.values:
                    plt.plot(xlim, [y, y], color=color, alpha=.3, lw=2,
                        zorder=1)

    if show_contour:
        try:
            sns.kdeplot(df.ix[:, :2].values, ax=ax)
        except ValueError:
            # Displaying a plot without a requested contour is better
            # than crashing.
            pass

            # This actually happens: with the 'skewed_numeric_5'
            # distribution in tests/test_plot_utils.py, we get:
            # seaborn/distributions.py:597: in kdeplot
            #     ax, **kwargs)
            # seaborn/distributions.py:380: in _bivariate_kdeplot
            #     cset = contour_func(xx, yy, z, n_levels, **kwargs)
            # matplotlib/axes/_axes.py:5333: in contour
            #     return mcontour.QuadContourSet(self, *args, **kwargs)
            # matplotlib/contour.py:1429: in __init__
            #     ContourSet.__init__(self, ax, *args, **kwargs)
            # matplotlib/contour.py:876: in __init__
            #     self._process_levels()
            # matplotlib/contour.py:1207: in _process_levels
            #     self.vmin = np.amin(self.levels)
            # numpy/core/fromnumeric.py:2224: in amin
            #     out=out, keepdims=keepdims)
            # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
            # a = array([], dtype=float64),
            # axis = None, out = None, keepdims = False
            #
            #     def _amin(a, axis=None, out=None, keepdims=False):
            # >       return umr_minimum(a, axis, None, out, keepdims)
            # E       ValueError: zero-size array to reduction operation
            #                     minimum which has no identity

    return ax


# No support for cyclic at this time
DO_PLOT_FUNC = dict()
DO_PLOT_FUNC[('categorical', 'categorical')] = do_heatmap
DO_PLOT_FUNC[('categorical', 'numerical')] = do_violinplot
DO_PLOT_FUNC[('numerical', 'categorical')] = do_violinplot
DO_PLOT_FUNC[('numerical', 'numerical')] = do_kdeplot


def do_pair_plot(plot_df, vartypes, **kwargs):
    # determine plot_types
    if kwargs.get('ax', None) is None:
        kwargs['ax'] = plt.gca()

    ax = DO_PLOT_FUNC[vartypes](plot_df, vartypes, **kwargs)
    return ax


def zmatrix(data_df, clustermap_kws=None, row_ordering=None,
        col_ordering=None, vmin=None, vmax=None):
    """Plots a clustermap from an ESTIMATE PAIRWISE query.

    Parameters
    ----------
    data_df : pandas.DataFrame
        The result of a PAIRWISE query in pandas.DataFrame.
    clustermap_kws : dict
        kwargs for seaborn.clustermap. See seaborn documentation. Of particular
        importance is the `pivot_kws` kwarg. `pivot_kws` is a dict with entries
        index, column, and values that let clustermap know how to reshape the
        data. If the query does not follow the standard ESTIMATE PAIRWISE
        output, it may be necessary to define `pivot_kws`.
        Other keywords here include vmin and vmax, linewidths, figsize, etc.
    row_ordering, col_ordering : list<int>
        Specify the order of labels on the x and y axis of the heatmap.
        To access the row and column indices from a clustermap object, use:
        clustermap.dendrogram_row.reordered_ind (for rows)
        clustermap.dendrogram_col.reordered_ind (for cols)
    vmin, vmax : float
        The minimum and maximum values of the colormap.

    Returns
    -------
    clustermap: seaborn.clustermap
    """
    if clustermap_kws is None:
        half_root_col = (data_df.shape[0] ** .5) / 2.0
        clustermap_kws = {'linewidths': 0.2, 'vmin': vmin, 'vmax': vmax,
                          'figsize': (half_root_col, .8 * half_root_col)}

    if clustermap_kws.get('pivot_kws', None) is None:
        # XXX: If the user doesnt tell us otherwise, we assume that this comes
        # fom a standard estimate pairwise query, which outputs columns
        # (table_id, col0, col1, value). The indices are indexed from the back
        # because it will also handle the no-table_id case
        data_df.columns = [' '*i for i in range(1, len(data_df.columns))] + \
            ['value']
        pivot_kws = {
            'index': data_df.columns[-3],
            'columns': data_df.columns[-2],
            'values': data_df.columns[-1],
        }
        clustermap_kws['pivot_kws'] = pivot_kws

    if clustermap_kws.get('cmap', None) is None:
        # Choose a soothing blue colormap
        clustermap_kws['cmap'] = 'BuGn'

    if row_ordering is not None and col_ordering is not None:
        index = clustermap_kws['pivot_kws']['index']
        columns = clustermap_kws['pivot_kws']['columns']
        values = clustermap_kws['pivot_kws']['values']
        df = data_df.pivot(index, columns, values)
        df = df.ix[:, col_ordering]
        df = df.ix[row_ordering, :]
        del clustermap_kws['pivot_kws']
        _fig, ax = plt.subplots()
        return (sns.heatmap(df, ax=ax, **clustermap_kws),
            row_ordering, col_ordering)
    else:
        return sns.clustermap(data_df, **clustermap_kws)


# TODO: bdb, and table_name should be optional arguments
def _pairplot(df, bdb=None, generator_name=None,
        show_contour=False, colorby=None, show_missing=False, show_full=False):
    """Plots the columns in data_df in a facet grid.

    Supports the following pairs:
    - categorical-categorical pairs are displayed as a heatmap
    - continuous-continuous pairs are displayed as a kdeplot
    - categorical-continuous pairs are displayed on a violin plot

    Parameters
    ----------
    df : pandas.DataFrame
        The input data---the result of a BQL/SQL query
    bdb : bayeslite.BayesDB (optional)
        The BayesDB object associated with `df`. Having the BayesDB object and
        the generator for the data allows pairplot to choose plot types.
    generator_name : str
        The name of generator associated with `df` and `bdb`.
    show_contour : bool
        If True, KDE contours are plotted on top of scatter plots
        and histograms.
    show_missing : bool
        If True, rows with one missing value are plotted as lines on scatter
        plots.
    colorby : str
        Name of a column to use to color data points in histograms and scatter
        plots.
    show_full : bool
        Show full pairwise plots, rather than only lower triangular plots.

    Returns
    -------
    figure : matplotlib.figure.Figure
        A num_columns by num_columns Gridspec of pairplot axes.

    Notes
    -----
    Support soon for ordered continuous combinations. It may be best
    to plot all ordered continuous pairs as heatmap.
    """
    # NOTE:Things to consider:
    # - String values are a possibility (categorical)
    # - who knows what the columns are named. What is the user selects columns
    #   as shortname?
    # - where to handle dropping NaNs? Missing values may be informative.

    data_df = df

    colors = None
    if colorby is not None:
        n_colorby = 0
        for colname in data_df.columns:
            # XXX: This is not guaranteed to work on all Unicode characters.
            if colorby.lower() == colname.lower():
                n_colorby += 1
                colorby = colname
        if n_colorby == 0:
            raise ValueError('colorby column, {}, not found.'.format(colorby))
        elif n_colorby > 1:
            raise ValueError('Multiple columns named, {}.'.format(colorby))

        dummy = data_df[colorby].dropna()
        dvals = np.sort(dummy.unique())
        ndvals = len(dvals)
        dval_type = get_bayesdb_col_type('colorby', dummy, bdb=bdb,
                                         generator_name=generator_name)
        if dval_type.lower() != 'categorical':
            raise ValueError('colorby columns must be categorical.')
        cmap = sns.color_palette("Set1", ndvals)
        colors = {}
        for val, color in zip(dvals, cmap):
            colors[val] = color

    all_varnames = [c for c in data_df.columns if c != colorby]
    n_vars = len(all_varnames)
    plt_grid = gridspec.GridSpec(n_vars, n_vars)
    figure = plt.figure()

    # if there is only one variable, just do a hist
    if n_vars == 1:
        ax = plt.gca()
        varname = data_df.columns[0]
        vartype = get_bayesdb_col_type(varname, data_df[varname], bdb=bdb,
                                       generator_name=generator_name)
        do_hist(data_df, dtype=vartype, ax=ax, bdb=bdb,
            generator_name=generator_name, colors=colors)
        if vartype == 'categorical':
            pass  # rotate_tick_labels(ax)
        return

    xmins = np.ones((n_vars, n_vars))*float('Inf')
    xmaxs = np.ones((n_vars, n_vars))*float('-Inf')
    ymins = np.ones((n_vars, n_vars))*float('Inf')
    ymaxs = np.ones((n_vars, n_vars))*float('-Inf')

    vartypes = []
    for varname in all_varnames:
        vartype = get_bayesdb_col_type(varname, data_df[varname], bdb=bdb,
            generator_name=generator_name)
        vartypes.append(vartype)

    # store each axes; reaccessing ax with plt.subplot(plt_grid[a,b]) may
    # overwrite the ax
    axes = [[] for _ in xrange(len(all_varnames))]
    for x_pos, var_name_x in enumerate(all_varnames):
        var_x_type = vartypes[x_pos]
        for y_pos, var_name_y in enumerate(all_varnames):
            var_y_type = vartypes[y_pos]

            ax = figure.add_subplot(plt_grid[y_pos, x_pos])
            axes[y_pos].append(ax)

            if x_pos == y_pos:
                varnames = [var_name_x]
                if colorby is not None:
                    varnames.append(colorby)
                ax = do_hist(data_df[varnames], dtype=var_x_type, ax=ax,
                    bdb=bdb, generator_name=generator_name, colors=colors)
            else:
                varnames = [var_name_x, var_name_y]
                vartypes_pair = (var_x_type, var_y_type,)
                if colorby is not None:
                    varnames.append(colorby)
                plot_df = prep_plot_df(data_df, varnames)
                ax = do_pair_plot(plot_df, vartypes_pair, ax=ax, bdb=bdb,
                    generator_name=generator_name,
                    show_contour=show_contour, show_missing=show_missing,
                    colors=colors)

                ymins[y_pos, x_pos] = ax.get_ylim()[0]
                ymaxs[y_pos, x_pos] = ax.get_ylim()[1]
                xmins[y_pos, x_pos] = ax.get_xlim()[0]
                xmaxs[y_pos, x_pos] = ax.get_xlim()[1]

            ax.set_xlabel(var_name_x, fontweight='bold')
            ax.set_ylabel(var_name_y, fontweight='bold')

    for x_pos in range(n_vars):
        for y_pos in range(n_vars):
            ax = axes[y_pos][x_pos]
            ax.set_xlim([np.min(xmins[:, x_pos]), np.max(xmaxs[:, x_pos])])
            if x_pos != y_pos:
                ax.set_ylim([np.min(ymins[y_pos, :]), np.max(ymaxs[y_pos, :])])
            if x_pos > 0:
                if x_pos == n_vars - 1:
                    ax.yaxis.tick_right()
                    ax.yaxis.set_label_position('right')
                else:
                    ax.set_ylabel('')
                    ax.set_yticklabels([])
            if y_pos < n_vars - 1:
                if y_pos == 0:
                    ax.xaxis.tick_top()
                    ax.xaxis.set_label_position('top')
                else:
                    ax.set_xlabel('')
                    ax.set_xticklabels([])
            else:
                if vartype[x_pos] == 'categorical':
                    rotate_tick_labels(ax)

    def fake_axis_ticks(ax_tl, ax_tn):
        atl, btl = ax_tl.get_ylim()
        atn, btn = ax_tn.get_ylim()
        tnticks = ax_tn.get_yticks()
        yrange_tn = (btn-atn)
        yrange_tl = (btl-atl)
        tntick_ratios = [(t-atn)/yrange_tn for t in tnticks]
        ax_tl.set_yticks([r*yrange_tl+atl for r in tntick_ratios])
        ax_tl.set_yticklabels(tnticks)

    # Fix the top-left histogram y-axis ticks and labels.
    if show_full:
        fake_axis_ticks(axes[0][0], axes[0][1])
    fake_axis_ticks(axes[-1][-1], axes[-1][0])

    if colorby is not None:
        legend = gen_collapsed_legend_from_dict(colors, title=colorby)
        legend.draggable()

    # tril by default by deleting upper diagonal axes.
    if not show_full:
        for y_pos in range(n_vars):
            for x_pos in range(y_pos+1, n_vars):
                figure.delaxes(axes[y_pos][x_pos])

    return figure


def comparative_hist(df, bdb=None, nbins=15, normed=False):
    """Plot a histogram

    Given a one-column pandas.DataFrame, df, plots a simple histogram. Given a
    two-column df plots the data in columns one separated by a a dummy variable
    assumed to be in column 2.

    Parameters
    ----------
    nbins : int
        Number of bins (bars)
    normed : bool
        If True, normalizes the the area of the histogram (or each
        sub-histogram if df has two columns) to 1.

    Returns
    -------
    figure: matplotlib.figure.Figure
    """
    df = df.dropna()

    vartype = get_bayesdb_col_type(df.columns[0], df[df.columns[0]], bdb=bdb)
    if vartype == 'categorical':
        values, labels, lookup = conv_categorical_vals_to_numeric(
            df[df.columns[0]])
        df.ix[:, 0] = values
        bins = len(labels)
        ticklabels = [0]*len(labels)
        for key, val in lookup.iteritems():
            ticklabels[val] = key
    else:
        a = min(df.ix[:, 0].values)
        b = max(df.ix[:, 0].values)
        support = b - a
        interval = support/nbins
        bins = np.linspace(a, b+interval, nbins)

    colorby = None
    if len(df.columns) > 1:
        if len(df.columns) > 2:
            raise ValueError('I do not know what to do with data with more '
                'than two columns.')
        colorby = df.columns[1]
        colorby_vals = df[colorby].unique()

    figure, ax = plt.subplots(tight_layout=False, facecolor='white')
    if colorby is None:
        ax.hist(df.ix[:, 0].values, bins=bins, color='#383838',
            edgecolor='none', normed=normed)
        plot_title = df.columns[0]
    else:
        colors = sns.color_palette('deep', len(colorby_vals))
        for color, cbv in zip(colors, colorby_vals):
            subdf = df[df[colorby] == cbv]
            ax.hist(subdf.ix[:, 0].values, bins=bins, color=color, alpha=.5,
                edgecolor='none', normed=normed, label=str(cbv))
        ax.legend(loc=0, title=colorby)
        plot_title = df.columns[0] + " by " + colorby

    if normed:
        plot_title += " (normalized)"

    ax.set_title(plot_title)
    ax.set_xlabel(df.columns[0])
    return figure



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import bayeslite
import bayeslite.core
import bayeslite.guess
from bayeslite.loggers import BqlLogger, logged_query
import bayeslite.metamodels.crosscat
import bdbcontrib
import bdbcontrib.plot_utils
from py_utils import helpsub

from collections import defaultdict
import crosscat
import crosscat.LocalEngine
import matplotlib.pyplot as plt
import re
import os
import pandas as pd
import sys
import traceback


class BqlRecipes(object):
  def __init__(self, name, csv_path=None, bdb_path=None, df=None, logger=None,
               session_capture_name=None):
    """A set of shortcuts for common ways to use BayesDB.

    name : str  REQUIRED.
        The name of dataset, should use letters and underscores only.
        This will also be used as a table name, and %t in queries will be
        replaced by this name.
    csv_path : str
        The path to a comma-separated values file. If specified, will be used
        to populate the bdb. It must exist and be both readable and non-empty.
    df : pandas.DataFrame
        If specified, these data will be used to populate the bdb, superseding
        any csv_path. It must not be empty.
    bdb_path : str
        If specified, store data and analysis at this location. If no other
        data source (csv or df) is specified, then it must already have been
        populated. If not specified, we will use a volatile in-memory bdb.
    logger : object
        Something on which we can call .info or .warn, by default a
        bayeslite.loggers.BqlLogger, but could be QuietLogger (only results),
        SilentLogger (nothing), IpyLogger, or LoggingLogger to use those
        modules, or anything else that implements the BqlLogger interface.
    session_capture_name : String
        None by default. If you want to send your session logs (perhaps
        including underlying data!) automatically to the Probabilistic
        Computing group for research purposes, possibly putting those logs and
        data into the public domain, then please set this value to a name that
        we might recognize you by. If the name is not self-evident, then please
        send that name in an email to bayesdb@mit.edu to help us connect your
        sessions to you.
    """
    assert re.match(r'\w+', name)
    assert df is not None or csv_path or bdb_path
    self.name = name
    self.generator_name = name + '_cc'
    self.csv_path = csv_path
    self.df = df
    self.bdb_path = bdb_path
    self.logger = BqlLogger() if logger is None else logger
    self.bdb = None
    self.status = None
    self.session_capture_name = session_capture_name
    self.initialize()

  def initialize(self):
    if self.bdb:
      return
    self.bdb = bayeslite.bayesdb_open(self.bdb_path)
    if not bayeslite.core.bayesdb_has_table(self.bdb, self.name):
      if self.df is not None:
        bayeslite.read_pandas.bayesdb_read_pandas_df(
          self.bdb, self.name, self.df, create=True, ifnotexists=True)
      elif self.csv_path:
        bayeslite.bayesdb_read_csv_file(
          self.bdb, self.name, self.csv_path,
          header=True, create=True, ifnotexists=True)
      else:
        raise ValueError("No data sources specified, and an empty bdb.")
    size = self.query('''SELECT COUNT(*) FROM %t''').ix(0, 0)
    assert 0 < size
    if "BAYESDB_WIZARD_MODE" in os.environ:
      old_wizmode = os.environ["BAYESDB_WIZARD_MODE"]
    else:
      old_wizmode = ""
    os.environ["BAYESDB_WIZARD_MODE"] = "1"
    self.query('''
        CREATE GENERATOR %g IF NOT EXISTS FOR %t USING crosscat( GUESS(*) )''')
    os.environ["BAYESDB_WIZARD_MODE"] = old_wizmode

  def check_representation(self):
    assert self.bdb, "Did you initialize?"

  def reset(self):
    self.check_representation()
    self.query('drop generator if exists %s' % self.generator_name)
    self.query('drop table if exists %s' % self.name)
    self.bdb = None
    self.initialize()

  @helpsub('bdbcontrib_describe', bdbcontrib.describe_generator_columns.__doc__)
  def quick_describe_columns(self):
    '''Wraps bdbcontrib.describe_generator_columns with bdb and generator name.

    bdbcontrib_describe'''
    self.check_representation()
    return bdbcontrib.describe_generator_columns(self.bdb, self.generator_name)

  help_for_query = (
      """Query the database. Use %t for the data table and %g for the generator.

      %t and %g work only with word boundaries. E.g., 'LIKE "%table%"' is fine.

      Returns a pandas.DataFrame with the results, rather than the cursor that
      the underlying bdb would return, so LIMIT your queries if you need to.
      """)

  @helpsub(r'help_for_query', help_for_query)
  def query(self, query_string, *bindings):
    '''Basic querying without session capture or reporting.

    help_for_query'''
    self.check_representation()
    query_string = re.sub(r'(^|(?<=\s))%t\b',
                          bayeslite.bql_quote_name(self.name),
                          re.sub(r'(^|(?<=\s))%g\b',
                                 bayeslite.bql_quote_name(self.generator_name),
                                 query_string))
    self.logger.info("BQL [%s] [%r]", query_string, bindings)
    with self.bdb.savepoint():
      try:
        res = self.bdb.execute(query_string, bindings)
        assert res is not None and res.description is not None
        self.logger.debug("BQL [%s] [%r] has returned a cursor." %
                          (query_string, bindings))
        df = bdbcontrib.cursor_to_df(res)
        self.logger.debug("BQL [%s] [%r] has created a dataframe." %
                            (query_string, bindings))
        return df
      except:
        self.logger.exception("")

  @helpsub(r'help_for_query', help_for_query)
  def q(self, query_string, *bindings):
    '''help_for_query'''
    with logged_query(query_string, bindings, name=self.session_capture_name):
      return self.query(query_string, *bindings)

  @helpsub('bdbcontrib_nullify_doc', bdbcontrib.nullify.__doc__)
  def nullify(self, value):
    """Wraps bdbcontrib.nullify by passing bdb and name.

    bdbcontrib_nullify_doc"""
    bdbcontrib.nullify(self.bdb, self.name, value)

  def analyze(self, models=100, minutes=0, iterations=0, checkpoint=0):
    '''Run analysis.

    models : integer
        The number of models bounds the accuracy of predictive probabilities.
        With ten models, then you get one decimal digit of interpretability,
        with a hundred models, you get two, and so on.
    minutes : integer
        How long you want to let it run.
    iterations : integer
        How many iterations to let it run.

    Returns:
        A report indicating how many models have seen how many iterations,
        and other info about model stability.
    '''
    self.check_representation()
    with logged_query(query_string='recipes.analyze',
                      name=self.session_capture_name):
      if models > 0:
        self.query('INITIALIZE %d MODELS IF NOT EXISTS FOR %s' %
              (models, self.generator_name))
        assert minutes == 0 or iterations == 0
      else:
        models = self.analysis_status().sum()
      if minutes > 0:
        if checkpoint == 0:
          checkpoint = max(1, int(minutes * models / 200))
        self.query('ANALYZE %s FOR %d MINUTES CHECKPOINT %d ITERATION WAIT' %
              (self.generator_name, minutes, checkpoint))
      elif iterations > 0:
        if checkpoint == 0:
          checkpoint = max(1, int(iterations / 20))
        self.query(
            '''ANALYZE %s FOR %d ITERATIONS CHECKPOINT %d ITERATION WAIT''' % (
              self.generator_name, iterations, checkpoint))
      else:
        raise NotImplementedError('No default analysis strategy yet. Please specify minutes or iterations.')
    # itrs = self.per_model_analysis_status()
    # models_with_fewest_iterations =
    #    itrs[itrs['iterations'] == itrs.min('index').head(0)[0]].index.tolist()
    # TODO(gremio): run each model with as many iterations as it needs to get
    # up to where it needs to get to, if that's larger?
    # Nope. Vikash said there's no reason to think that's a good idea. Perhaps
    # even better to have some young models mixed in with the old ones.
    # I still think we should make some recommendation that scales for what
    # "the right thing" is, where that's something that at least isn't known to
    # suck.

    return self.analysis_status()

  # TODO: remove import plot_utils from the __init__.py file -- make it empty.
  # If you want to use it, you should import bdbcontrib.foo.

  def per_model_analysis_status(self):
    """Return the number of iterations for each model."""
    # XXX Move this to bdbcontrib/src/bql_utils.py ?
    try:
      return self.query('''SELECT iterations FROM bayesdb_generator_model
                           WHERE generator_id = (
                            SELECT id FROM bayesdb_generator WHERE name = ?)''',
                        self.generator_name)
    except ValueError:
      # Because, e.g. there is no generator yet, for an empty db.
      return None

  def analysis_status(self):
    """Return the count of models for each number of iterations run."""
    itrs = self.per_model_analysis_status()
    if itrs is None or len(itrs) == 0:
      emt = pd.DataFrame(columns=['count of models'])
      emt.index.name = 'iterations'
      return emt
    vcs = pd.DataFrame(itrs['iterations'].value_counts())
    vcs.index.name = 'iterations'
    vcs.columns = ['count of models']
    self.status = vcs
    return vcs

  @helpsub('bdbcontrib_pairplot', bdbcontrib.plot_utils.pairplot.__doc__)
  def pairplot(self, cols, plotfile=None, colorby=None, **kwargs):
    """Wrap bdbcontrib.plot_utils.pairplot to show the given columns.

    Specifies bdb, query with the given columns, and generator_name:
    bdbcontrib_pairplot
    """
    if len(cols) < 1:
        raise ValueError('Pairplot at least one variable.')
    qcols = cols if colorby is None else set(cols + [colorby])
    query_columns = '''"%s"''' % '''", "'''.join(qcols)
    with logged_query(query_string='pairplot cols=?', bindings=(query_columns,),
                      name=self.session_capture_name):
      self.logger.plot(plotfile,
                       bdbcontrib.pairplot(self.bdb, '''SELECT %s FROM %s''' %
                                             (query_columns, self.name),
                                           generator_name=self.generator_name,
                                           colorby=colorby,
                                           **kwargs))

  def heatmap(self, deps, selectors=None, plotfile=None, **kwargs):
    '''Show heatmaps for the given dependencies

    Parameters
    ----------
    deps : pandas.DataFrame(columns=['generator_id', 'name0', 'name1', 'value'])
        The result of a .q('ESTIMATE ... PAIRWISE ...')
        E.g., DEPENDENCE PROBABILITY, MUTUAL INFORMATION, COVARIANCE, etc.

    selectors : {str: lambda name --> bool}
        Rather than plot the full NxN matrix all together, make separate plots
        for each combination of these selectors, plotting them in sequence.
        If selectors are specified, the actual selector functions are values of
        a dict, and the keys are their names, for purposes of plot legends and
        filenames.
        E.g.,
          {'A-E': lambda x: bool(re.search(r'^[a-eA-E]', x[0])),
           'F-O': lambda x: bool(re.search(r'^[f-oF-O]', x[0])),
           'P-Z': lambda x: bool(re.search(r'^[p-zP-Z]', x[0]))}

    plotfile : str
        If a plotfile is specified, savefig to that file. If selectors are also
        specified, savefig to name1.name2.plotfile.

    **kwargs : dict
        Passed to zmatrix: vmin, vmax, row_ordering, col_ordering
    '''
    with logged_query(query_string='heatmap(deps, selectors)',
                      bindings=(str(deps), repr(selectors)),
                      name=self.session_capture_name):
      hmap = plt.figure()
      if selectors is None:
        cmap = bdbcontrib.plot_utils.heatmap(self.bdb, df=deps, **kwargs)
        self.logger.plot(plotfile, cmap)
      else:
        selfns = [selectors[k] for k in sorted(selectors.keys())]
        reverse = dict([(v, k) for (k, v) in selectors.items()])
        for (cmap, sel1, sel2) in bdbcontrib.plot_utils.selected_heatmaps(
             self.bdb, df=deps, selectors=selfns, **kwargs):
          self.logger.plot("%s.%s.%s" % (
              reverse[sel1], reverse[sel2], plotfile),
                           cmap)
      return hmap

  def quick_explore_cols(self, cols, nsimilar=20, plotfile='explore_cols'):
    """Show dependence probabilities and neighborhoods based on those.

    cols: list of strings
        At least two column names to look at dependence probabilities of,
        and to explore neighborhoods of.
    nsimilar: positive integer
        The size of the neighborhood to explore.
    plotfile: string pathname
        Where to save plots, if not displaying them on console.
    """
    if len(cols) < 2:
      raise ValueError('Need to explore at least two columns.')
    with logged_query(query_string='quick_explore_cols', bindings=(cols,),
                      name=self.session_capture_name):
      self.pairplot(cols)
      query_columns = '''"%s"''' % '''", "'''.join(cols)
      deps = self.query('''ESTIMATE DEPENDENCE PROBABILITY
                           FROM PAIRWISE COLUMNS OF %s
                           FOR %s;''' % (self.generator_name, query_columns))
      deps.columns = ['genid', 'name0', 'name1', 'value']
      self.heatmap(deps, plotfile=plotfile)
      deps.columns = ['genid', 'name0', 'name1', 'value']
      triangle = deps[deps['name0'] < deps['name1']]
      triangle = triangle.sort_values(ascending=False, by=['value'])
      self.logger.result("Pairwise dependence probability for: %s\n%s\n\n",
                         query_columns, triangle)

      for col in cols:
        neighborhood = self.query(
        '''ESTIMATE *, DEPENDENCE PROBABILITY WITH "%s"
           AS "Probability of Dependence with %s"
           FROM COLUMNS OF %s
           ORDER BY "Probability of Dependence with %s"
           DESC LIMIT %d;'''
           % (col, col, self.generator_name, col, nsimilar))
        neighbor_columns = ('''"%s"''' %
                            '''", "'''.join(neighborhood["name"].tolist()))
        deps = self.query('''ESTIMATE DEPENDENCE PROBABILITY
            FROM PAIRWISE COLUMNS OF %s
            FOR %s;''' % (self.generator_name, neighbor_columns))
        deps.columns = ['genid', 'name0', 'name1', 'value']
        self.heatmap(deps, plotfile=(plotfile + "-" + col))
        self.logger.result("Pairwise dependence probability of %s with its " +
                           "strongest dependents:\n%s\n\n", col, neighborhood)

  def column_type(self, col):
    """The statistical type of the given column in the current model."""
    descriptions = self.quick_describe_columns()
    return descriptions[descriptions['name'] == col]['stattype'].iloc[0]

  def sql_tracing(self, turn_on=True):
    """Trace underlying SQL, for debugging."""
    # Always turn off:
    self.bdb.sql_untrace(self.bdb.sql_tracer)
    if turn_on:
      printer = lambda query, bindings: self.logger.info(
          "query: [%s] bindings: [%s]\n\n", query, bindings)
      self.bdb.sql_trace(printer)

  def quick_similar_rows(self, identify_row_by, nsimilar=10):
    """Explore rows similar to the identified one.

    identify_row_by : dict
        Dictionary of column names to their values. These will be turned into
        a WHERE clause in BQL, and must identify one unique row.
    nsimilar : positive integer
        The number of similar rows to retrieve.
    """
    with logged_query(query_string='quick_similar_rows(id_by, n)',
                      bindings=(identify_row_by, nsimilar),
                      name=self.session_capture_name):
      import hashlib
      table_name = 'tmptbl_' + hashlib.md5('\x00'.join(
          [repr(identify_row_by), str(self.status)])).hexdigest()
      column_name = 'similarity_to_' + "__".join(
          re.sub(r'\W', '_', str(val)) for val in identify_row_by.values())
      query_params = []
      query_columns = []
      for k, v in identify_row_by.iteritems():
        query_columns.append('''%s = ? ''' % bayeslite.bql_quote_name(k))
        query_params.append(v)
      query_attrs = ' and '.join(query_columns)

      with self.bdb.savepoint():
        row_exists = self.query('SELECT COUNT(*) FROM %s WHERE %s;' %
                                (self.name, query_attrs))
        if row_exists.ix[0][0] != 1:
          raise NotImplementedError(
              'identify_row_by found %d rows instead of exactly 1 in %s.' %
              (row_exists.ix[0][0], self.csv_path))
        creation_query = ('''CREATE TEMP TABLE IF NOT EXISTS %s AS ESTIMATE *,
                             SIMILARITY TO (%s) AS %s FROM %%g LIMIT %d;''' %
                          (table_name, query_attrs, column_name, nsimilar))
        self.query(creation_query, query_params)
        result = self.query('''SELECT * FROM %s ORDER BY %s DESC;''' %
                            (table_name, column_name))
      return result

@helpsub('__init__', BqlRecipes.__init__.__doc__)
def quickstart(*args, **kwargs):
  '''__init__'''
  return BqlRecipes(*args, **kwargs)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt

import mock
import numpy as np
import re
import os
import pandas as pd
from io import BytesIO

import bayeslite
import bdbcontrib

from bayeslite.read_csv import bayesdb_read_csv
from bdbcontrib.bql_utils import cursor_to_df
from bdbcontrib.plot_utils import _pairplot

def dataset(num_rows=400):
    '''Column names give rough type, numbers show which are related.'''
    df = pd.DataFrame()
    alphabet = ['A', 'B', 'C', 'D', 'E']
    dist_of_five = np.random.choice(range(5), num_rows,
                                    p=np.array([1, .4, .3, .2, .1]) / 2.)
    df['floats_1'] = [np.random.randn() + x for x in dist_of_five]
    df['categorical_1'] = [alphabet[i] for i in dist_of_five]

    dist_of_five = np.random.choice(range(5), num_rows,
                                    p=np.array([1, .4, .3, .2, .1]) / 2.)
    df['categorical_2'] = [alphabet[i] for i in dist_of_five]

    few_ints = np.random.choice(range(4), num_rows, p=[.4, .3, .2, .1])
    df['few_ints_3'] = few_ints
    df['floats_3'] = [(np.random.randn() - 2 * x) / (1 + x) for x in few_ints]

    df['many_ints_4'] = np.random.randn(num_rows)

    # Need >= 75% to be zero to trigger inter-quartile-range == 0.
    numeric_5 = np.random.choice(
        range(4), num_rows, p=[200./256, 40./256, 10./256, 6./256])
    # Need >= 30 unique to be numeric instead of categorical,
    # so perturb the non-zeros:
    df['skewed_numeric_5'] = [ i * np.random.random() for i in numeric_5 ]
    csv_data = BytesIO()
    df.to_csv(csv_data, header=True, index_label='index')
    return (df, csv_data)

def prepare():
    (df, csv_str) = dataset()
    os.environ['BAYESDB_WIZARD_MODE']='1'
    bdb = bayeslite.bayesdb_open()
    # XXX Do we not have a bayesdb_read_df ?
    bayesdb_read_csv(bdb, 'plottest', flush(csv_str),
                     header=True, create=True)
    bdb.execute('''
        create generator plottest_cc for plottest using crosscat(guess(*))
    ''')

    # do a plot where a some sub-violins are removed
    _remove_violin_bql = """
        DELETE FROM plottest
            WHERE categorical_1 = "B"
                AND (few_ints_3 = 2 OR few_ints_3 = 1);
    """
    cursor = bdb.execute('SELECT * FROM plottest')
    df = cursor_to_df(cursor)
    return (df, bdb)

def do(prepped, location, **kwargs):
    (df, bdb) = prepped
    plt.figure(tight_layout=True, facecolor='white')
    _pairplot(df, bdb=bdb, generator_name='plottest_cc',
              show_full=False, **kwargs)
    plt.savefig(location)


from PIL import Image
def has_nontrivial_contents_over_white_background(imgfile):
    img = Image.open(imgfile)
    pix = img.load()
    hist = img.histogram()
    # White background means corners are likely white:
    assert pix[0, 0] == (255, 255, 255, 255)
    assert pix[img.size[0] - 1, img.size[1] - 1] == (255, 255, 255, 255)
    white = hist[-1]
    total = sum(hist)
    # Total white should be between 10 and 90% of the image.
    assert total * .1 < white
    assert white < total * .9
    # Total number of different intensities used should be between 10 and 90%
    # of the available intensities.
    zeros = len([intensity for intensity in hist if intensity == 0])
    assert len(hist) * .1 < zeros
    assert zeros < len(hist) * .9
    return True

def flush(f):
    return BytesIO(f.getvalue())

def test_pairplot_smoke():
    ans = prepare()
    f = BytesIO()
    do(ans, f, colorby='categorical_2', show_contour=False)
    assert has_nontrivial_contents_over_white_background(flush(f))
    f = BytesIO()
    do(ans, f, colorby='categorical_2', show_contour=True)
    assert has_nontrivial_contents_over_white_background(flush(f))
    f = BytesIO()
    do(ans, f, show_contour=False)
    assert has_nontrivial_contents_over_white_background(flush(f))

def test_one_variable():
    (df, bdb) = prepare()
    for var in ['categorical_1', 'few_ints_3', 'floats_3', 'many_ints_4',
                'skewed_numeric_5']:
      cursor = bdb.execute('SELECT %s FROM plottest' % (var,))
      df = cursor_to_df(cursor)
      f = BytesIO()
      do((df, bdb), f, show_contour=False)
      assert has_nontrivial_contents_over_white_background(flush(f))
      cursor = bdb.execute('SELECT %s, categorical_2 FROM plottest' % (var,))
      df = cursor_to_df(cursor)
      f = BytesIO()
      do((df, bdb), f, colorby='categorical_2', show_contour=False)
      assert has_nontrivial_contents_over_white_background(flush(f))
      f = BytesIO()
      do((df, bdb), f, colorby='categorical_2', show_contour=True)
      assert has_nontrivial_contents_over_white_background(flush(f))

def test_selected_heatmaps():
    (unused_df, bdb) = prepare()
    s_ae = lambda x: bool(re.search(r'^[a-eA-E]', x[0]))
    s_fw = lambda x: bool(re.search(r'^[f-wF-W]', x[0]))
    s_xz = lambda x: bool(re.search(r'^[x-zX-Z]', x[0]))
    sels = [s_ae, s_fw, s_xz]
    deps = pd.DataFrame([[0, 'a', 'a', 1],
                         [0, 'a', 'b', 1],
                         [0, 'a', 'f', 1],
                         [0, 'b', 'a', 1],
                         [0, 'b', 'b', 1],
                         [0, 'b', 'f', 1],
                         [0, 'f', 'a', 1],
                         [0, 'f', 'b', 1],
                         [0, 'f', 'f', 1]])
    seen = []
    with mock.patch('bdbcontrib.plot_utils.zmatrix', return_value=42):
        for (plot, s0, s1) in bdbcontrib.plot_utils.selected_heatmaps(
                bdb, selectors=sels, df=deps):
            assert 42 == plot
            seen.append((s0, s1))
    assert 4 == len(seen)
    assert (s_ae, s_ae) in seen
    assert (s_ae, s_fw) in seen
    assert (s_fw, s_ae) in seen
    assert (s_fw, s_fw) in seen


def get_plot_text(obj):
    texts = []
    stack = [obj]
    while stack:
        item = stack.pop()
        d = set(dir(item))
        if 'get_children' in d:
            stack += item.get_children()
        if 'get_text' in d:
            texts.append(item.get_text())
    return texts

def test_gen_collapsed_legend_from_dict():
    hl_colors_dict = {'roses': 'red',
                      'violets': 'blue',
                      'lilies': 'white',
                      'poppies': 'red'}
    # http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.legend
    loc = 10  # center
    title = 'Doggerel!'
    fontsize = 'medium'
    leg = bdbcontrib.plot_utils.gen_collapsed_legend_from_dict(
        hl_colors_dict, loc, title, fontsize)
    texts = get_plot_text(leg)
    assert 'Doggerel!' in texts
    assert 'violets' in texts
    assert 'lilies' in texts
    assert ('roses, poppies' in texts or 'poppies, roses' in texts)


def main():
    ans = prepare()
    do(ans, 'fig0.png', colorby='categorical_2', show_contour=False)
    do(ans, 'fig1.png', colorby='categorical_2', show_contour=True)
    do(ans, 'fig2.png', show_contour=False)
    print "Figures saved in 'fig0.png', 'fig1.png', 'fig2.png'"
    assert os.path.exists('fig0.png')
    assert os.path.exists('fig1.png')
    assert os.path.exists('fig2.png')

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

try:
    from setuptools import setup
    from setuptools.command.test import test as TestCommand
except ImportError:
    from distutils.core import setup
    from distutils.cmd import Command
    class TestCommand(Command):
        user_options = []
        def initialize_options(self): pass
        def finalize_options(self): pass
        def run(self): self.run_tests()

with open('VERSION', 'rU') as f:
    version = f.readline().strip()

from distutils.command.install import INSTALL_SCHEMES
for scheme in INSTALL_SCHEMES.values():
    scheme['data'] = scheme['purelib']
example_files = {}
import os
import re
with open('MANIFEST.in', 'r') as manifest:
    for line in manifest:
        line = line.strip()
        if line.startswith('include examples'):
            line = line[len("include "):]
            dirname = os.path.join('bdbcontrib', os.path.dirname(line))
            if dirname not in example_files:
                example_files[dirname] = []
            example_files[dirname].append(line)

# Append the Git commit id if this is a development version.
if version.endswith('+'):
    tag = 'v' + version[:-1]
    try:
        import subprocess
        desc = subprocess.check_output([
            'git', 'describe', '--dirty', '--match', tag,
        ])
    except Exception:
        version += 'unknown'
    else:
        assert desc.startswith(tag)
        import re
        match = re.match(r'v([^-]*)-([0-9]+)-(.*)$', desc)
        if match is None:       # paranoia
            version += 'unknown'
        else:
            ver, rev, local = match.groups()
            version = '%s.post%s+%s' % (ver, rev, local.replace('-', '.'))
            assert '-' not in version

try:
    with open('src/version.py', 'rU') as f:
        version_old = f.readlines()
except IOError:
    version_old = None
version_new = ['__version__ = %s\n' % (repr(version),)]
if version_old != version_new:
    with open('src/version.py', 'w') as f:
        f.writelines(version_new)

# XXX Several horrible kludges here to make `python setup.py test' work:
#
# - Standard setputools test command searches for unittest, not
#   pytest.
#
# - pytest suggested copypasta assumes . works in sys.path; we
#   deliberately make . not work in sys.path and require ./build/lib
#   instead, in order to force a clean build.
#
# - Must set PYTHONPATH too for shell tests, which fork and exec a
#   subprocess which inherits PYTHONPATH but not sys.path.
#
# - build command's build_lib variable is relative to source
#   directory, so we must assume os.getcwd() gives that.
#
# (The shell test issues are irrelevant for the moment in bdbcontrib,
# but there's no harm in setting PYTHONPATH anyway.)
class cmd_pytest(TestCommand):
    def __init__(self, *args, **kwargs):
        TestCommand.__init__(self, *args, **kwargs)
        self.test_suite = 'tests'
        self.build_lib = None
    def finalize_options(self):
        TestCommand.finalize_options(self)
        # self.build_lib = ...
        self.set_undefined_options('build', ('build_lib', 'build_lib'))
    def run_tests(self):
        import pytest
        import os
        import os.path
        import sys
        sys.path = [os.path.join(os.getcwd(), self.build_lib)] + sys.path
        os.environ['BAYESDB_WIZARD_MODE'] = '1'
        os.environ['BAYESDB_DISABLE_VERSION_CHECK'] = '1'
        os.environ['PYTHONPATH'] = ':'.join(sys.path)
        sys.exit(pytest.main(['tests']))

setup(
    name='bdbcontrib',
    version=version,
    description='Hodgepodge library of extras for bayeslite',
    url='http://probcomp.csail.mit.edu/bayesdb',
    author='MIT Probabilistic Computing Project',
    author_email='bayesdb@mit.edu',
    license='Apache License, Version 2.0',
    install_requires=[
        'bayeslite>=0.1.6',
        'ipython[notebook]>=3',
        'markdown2',
        'matplotlib==1.4.3',  # See bdbcontrib/issues/94
        'numpy==1.8.2',       # See bdbcontrib/issues/94
        'numpydoc',
        'pandas>=0.17.0',
        'seaborn>=0.6',
        'scipy==0.15.1',      # See bdbcontrib/issues/94
        'sklearn',
        'sklearn-pandas',
        'sphinx',
        'tornado>=4.0',
    ],
    tests_require=[
        'flaky',  # for bdb-experiments
        'mock',
        'pillow',
        'pytest',
    ],
    packages=[
        'bdbcontrib',
        'bdbcontrib.predictors',
        'bdbcontrib.metamodels',
        'bdbcontrib.experiments',
    ],
    package_dir={
        'bdbcontrib': 'src',
    },
    data_files=example_files.items(),
    scripts=[
        'scripts/bayesdb-demo',
    ],
    cmdclass={
        'test': cmd_pytest,
    },
)
